// -*- lsst-c++ -*-

/*
 * LSST Data Management System
 * Copyright 2008, 2009, 2010 LSST Corporation.
 *
 * This product includes software developed by the
 * LSST Project (http://www.lsst.org/).
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the LSST License Statement and
 * the GNU General Public License along with this program.  If not,
 * see <http://www.lsstcorp.org/LegalNotices/>.
 */

/*
 * Support for 2-D images
 *
 * This file contains the basic 2-d image support for LSST
 */
#ifndef LSST_AFW_IMAGE_IMAGEBASE_H
#define LSST_AFW_IMAGE_IMAGEBASE_H

#include <string>
#include <utility>
#include <functional>

#include <memory>

#include "lsst/geom.h"
#include "lsst/afw/image/lsstGil.h"
#include "lsst/afw/image/ImageUtils.h"
#include "lsst/afw/math/Function.h"
#include "lsst/pex/exceptions.h"
#include "ndarray.h"

namespace lsst {
namespace afw {

namespace fits {
class Fits;
class MemFileManager;
struct ImageWriteOptions;
}  // namespace fits

namespace image {
namespace detail {
//
// Traits for image types
//
/// Base %image tag
struct basic_tag {};
/// tag for an Image
struct Image_tag : public basic_tag {};
/// traits class for image categories
template <typename ImageT>
struct image_traits {
    using image_category = typename ImageT::image_category;
};
//
std::string const wcsNameForXY0 = "A";  // the name of the WCS to use to save (X0, Y0) to FITS files; e.g. "A"
}  // namespace detail

/// A class used to request that array accesses be checked
class CheckIndices {
public:
    explicit CheckIndices(bool check = true) : _check(check) {}
    operator bool() const { return _check; }

private:
    bool _check;
};

/// metafunction to extract reference type from PixelT
template <typename PixelT>
struct Reference {
    using type = typename boost::gil::channel_traits<PixelT>::reference;  ///< reference type
};
/// metafunction to extract const reference type from PixelT
template <typename PixelT>
struct ConstReference {
    using type = typename boost::gil::channel_traits<PixelT>::const_reference;  ///< const reference type
};

enum ImageOrigin { PARENT, LOCAL };

/// The base class for all %image classed (Image, Mask, MaskedImage, ...)
//
// You are not expected to use this class directly in your own code; use one of the
// specialised subclasses
//
template <typename PixelT>
class ImageBase {
private:
    using _view_t = typename lsst::afw::image::detail::types_traits<PixelT>::view_t;
    using _const_view_t = typename lsst::afw::image::detail::types_traits<PixelT>::const_view_t;

    using Manager = ndarray::Manager;

public:
    using image_category = detail::basic_tag;  ///< trait class to identify type of %image

    /// A single Pixel of the same type as those in the ImageBase
    using SinglePixel = PixelT;
    /// A pixel in this ImageBase
    using Pixel = PixelT;
    /// A Reference to a PixelT
    using PixelReference = typename Reference<PixelT>::type;
    /// A ConstReference to a PixelT
    using PixelConstReference = typename ConstReference<PixelT>::type;
    /// An xy_locator
    using xy_locator = typename _view_t::xy_locator;
    /// A const_xy_locator
    using const_xy_locator = typename _view_t::xy_locator::const_t;
    /// An STL compliant iterator
    using iterator = typename _view_t::iterator;
    /// An STL compliant const iterator
    using const_iterator = typename _const_view_t::iterator;
    /// An STL compliant reverse iterator
    using reverse_iterator = typename _view_t::reverse_iterator;
    /// An STL compliant const reverse iterator
    using const_reverse_iterator = typename _const_view_t::reverse_iterator;
    /// An iterator for traversing the pixels in a row
    using x_iterator = typename _view_t::x_iterator;
    /** A fast STL compliant iterator for contiguous images
     * N.b. The order of pixel access is undefined
     */
    using fast_iterator = x_iterator;
    /// An iterator for traversing the pixels in a row, created from an xy_locator
    using xy_x_iterator = typename _view_t::x_iterator;
    /// A const iterator for traversing the pixels in a row
    using const_x_iterator = typename _const_view_t::x_iterator;
    /// An iterator for traversing the pixels in a column
    using y_iterator = typename _view_t::y_iterator;
    /// An iterator for traversing the pixels in a row, created from an xy_locator
    using xy_y_iterator = typename _view_t::y_iterator;
    /// A const iterator for traversing the pixels in a column
    using const_y_iterator = typename _const_view_t::y_iterator;
    /// A mutable ndarray representation of the image
    using Array = typename ndarray::Array<PixelT, 2, 1>;
    /// An immutable ndarray representation of the image
    using ConstArray = typename ndarray::Array<const PixelT, 2, 1>;

    template <typename OtherPixelT>
    friend class ImageBase;  // needed by generalised copy constructors

    /// Convert a type to our SinglePixel type
    template <typename SinglePixelT>
    static SinglePixel PixelCast(SinglePixelT rhs) {
        return SinglePixel(rhs);
    }
    //
    // DecoratedImage needs enough access to ImageBase to read data from disk; we might be able to design
    // around this
    //
    template <typename>
    friend class DecoratedImage;
    template <typename, typename, typename>
    friend class MaskedImage;
    /**
     * Allocator Constructor
     *
     * allocate a new image with the specified dimensions.
     * Sets origin at (0,0)
     */
    explicit ImageBase(const lsst::geom::Extent2I& dimensions = lsst::geom::Extent2I());
    /**
     * Allocator Constructor
     *
     * allocate a new image with the specified dimensions and origin
     */
    explicit ImageBase(const lsst::geom::Box2I& bbox);
    /**
     * Copy constructor.
     *
     * @param src Right-hand-side %image
     * @param deep If false, new ImageBase shares storage with `src`; if true make a new, standalone,
     * ImageBase
     *
     * @note Unless `deep` is `true`, the new %image will share the old %image's pixels;
     * this may not be what you want.  See also assign(rhs) to copy pixels between Image%s
     */
    ImageBase(const ImageBase& src, const bool deep = false);
    ImageBase(ImageBase&& src);
    /**
     * Copy constructor to make a copy of part of an %image.
     *
     * The bbox ignores X0/Y0 if origin == LOCAL, and uses it if origin == PARENT.
     *
     * @param src Right-hand-side %image
     * @param bbox Specify desired region
     * @param origin Specify the coordinate system of the bbox
     * @param deep If false, new ImageBase shares storage with `src`; if true make a new, standalone,
     * ImageBase
     *
     * @note Unless `deep` is `true`, the new %image will share the old %image's pixels;
     * this is probably what you want
     */
    explicit ImageBase(const ImageBase& src, const lsst::geom::Box2I& bbox, const ImageOrigin origin = PARENT,
                       const bool deep = false);
    /**
     * generalised copy constructor
     *
     * defined here in the header so that the compiler can instantiate N(N-1) conversions between N
     * ImageBase types.
     */
    template <typename OtherPixelT>
    ImageBase(const ImageBase<OtherPixelT>& rhs, const bool deep) {
        if (!deep) {
            throw LSST_EXCEPT(lsst::pex::exceptions::InvalidParameterError,
                              "Only deep copies are permitted for ImageBases with different pixel types");
        }

        ImageBase<PixelT> tmp(rhs.getBBox());
        copy_and_convert_pixels(rhs._gilView, tmp._gilView);  // from boost::gil

        using std::swap;               // See Meyers, Effective C++, Item 25
        ImageBase<PixelT>::swap(tmp);  // See Meyers, Effective C++, Items 11 and 43
    }

    /**
     *  Construction from ndarray::Array and NumPy.
     *
     *  @note ndarray and NumPy indexes are ordered (y,x), but Image indices are ordered (x,y).
     *
     *  Unless deep is true, the new image will share memory with the array if the the
     *  dimension is contiguous in memory.  If the last dimension is not contiguous, the array
     *  will be deep-copied in Python, but the constructor will fail to compile in pure C++.
     */
    explicit ImageBase(Array const& array, bool deep = false,
                       lsst::geom::Point2I const& xy0 = lsst::geom::Point2I());

    virtual ~ImageBase() = default;
    /** Shallow assignment operator.
     *
     * @note that this has the effect of making the lhs share pixels with the rhs which may
     * not be what you intended;  to copy the pixels, use assign(rhs)
     *
     * @note this behaviour is required to make the swig interface work, otherwise I'd
     * declare this function private
     */
    ImageBase& operator=(const ImageBase& rhs);
    ImageBase& operator=(ImageBase&& rhs);
    /// Set the %image's pixels to rhs
    ImageBase& operator=(const PixelT rhs);

    /**
     * Copy pixels from another image to a specified subregion of this image.
     *
     * @param[in] rhs  source image whose pixels are to be copied into this image (the destination)
     * @param[in] bbox  subregion of this image to set; if empty (the default) then all pixels are set
     * @param[in] origin  origin of bbox: if PARENT then the lower left pixel of this image is at xy0
     *                    if LOCAL then the lower left pixel of this image is at 0,0
     *
     * @throws lsst::pex::exceptions::LengthError if the dimensions of rhs and the specified subregion of
     * this image do not match.
     */
    void assign(ImageBase const& rhs, lsst::geom::Box2I const& bbox = lsst::geom::Box2I(),
                ImageOrigin origin = PARENT);

    // TODO: should deprecate these, but the replacement should take two ints
    // rather than a Point2I, to keep C++ loops simple
    /// Return a reference to the pixel `(x, y)` in LOCAL coordinates
    PixelReference operator()(int x, int y);
    /// Return a reference to the pixel `(x, y)` in LOCAL coordinates with bounds checking
    PixelReference operator()(int x, int y, CheckIndices const&);
    /// Return a const reference to the pixel `(x, y)` in LOCAL coordinates
    PixelConstReference operator()(int x, int y) const;
    /// Return a const reference to the pixel `(x, y)` in LOCAL coordinates with bounds checking
    PixelConstReference operator()(int x, int y, CheckIndices const&) const;

    /// Return a reference to a single pixel (with no bounds check).
    PixelReference get(lsst::geom::Point2I const& index, ImageOrigin origin);

    /// Return a const reference to a single pixel (with no bounds check).
    PixelConstReference get(lsst::geom::Point2I const& index, ImageOrigin origin) const;

    /// Return a reference to a single pixel in PARENT coordinates (with no bounds check).
    PixelReference operator[](lsst::geom::Point2I const& index) { return get(index, PARENT); }

    /// Return a reference to a single pixel in PARENT coordinates (with no bounds check).
    PixelConstReference operator[](lsst::geom::Point2I const& index) const { return get(index, PARENT); }

    /// Return the number of columns in the %image
    int getWidth() const { return _gilView.width(); }
    /// Return the number of rows in the %image
    int getHeight() const { return _gilView.height(); }
    /// Return the area of the %image
    int getArea() const { return getWidth()*getHeight(); }
    /**
     * Return the %image's column-origin
     *
     * This will usually be 0 except for images created using the
     * `ImageBase(fileName, hdu, BBox, mode)` ctor or `ImageBase(ImageBase, BBox)` cctor
     * The origin can be reset with `setXY0`
     */
    int getX0() const { return _origin.getX(); }
    /**
     * Return the %image's row-origin
     *
     * This will usually be 0 except for images created using the
     * `ImageBase(fileName, hdu, BBox, mode)` ctor or `ImageBase(ImageBase, BBox)` cctor
     * The origin can be reset with `setXY0`
     */
    int getY0() const { return _origin.getY(); }

    /**
     * Return the %image's origin
     *
     * This will usually be (0, 0) except for images created using the
     * `ImageBase(fileName, hdu, BBox, mode)` ctor or `ImageBase(ImageBase, BBox)` cctor
     * The origin can be reset with `setXY0`
     */
    lsst::geom::Point2I getXY0() const { return _origin; }

    /**
     * Convert image position to index (nearest integer and fractional parts)
     *
     * @returns std::pair(nearest integer index, fractional part)
     */
    std::pair<int, double> positionToIndex(
            double const pos,                ///< image position
            lsst::afw::image::xOrY const xy  ///< Is this a column or row coordinate?
            ) const {
        double const fullIndex = pos - PixelZeroPos - (xy == X ? getX0() : getY0());
        int const roundedIndex = static_cast<int>(fullIndex + 0.5);
        double const residual = fullIndex - roundedIndex;
        return std::pair<int, double>(roundedIndex, residual);
    }

    /**
     * Convert image index to image position
     *
     * The LSST indexing convention is:
     * * the index of the bottom left pixel is 0,0
     * * the position of the center of the bottom left pixel is PixelZeroPos, PixelZeroPos
     *
     * @returns image position
     */
    inline double indexToPosition(double ind,                      ///< image index
                                  lsst::afw::image::xOrY const xy  ///< Is this a column or row coordinate?
                                  ) const {
        return ind + PixelZeroPos + (xy == X ? getX0() : getY0());
    }

    /// Return the %image's size;  useful for passing to constructors
    lsst::geom::Extent2I getDimensions() const { return lsst::geom::Extent2I(getWidth(), getHeight()); }

    void swap(ImageBase& rhs);

    Array getArray();
    ConstArray getArray() const;
    //
    // Iterators and Locators
    //
    /** Return an STL compliant iterator to the start of the %image
     *
     * Note that this isn't especially efficient; see @link imageIterators Image Iterators @endlink for
     * a discussion
     */
    iterator begin() const;
    /// Return an STL compliant iterator to the end of the %image
    iterator end() const;
    /// Return an STL compliant reverse iterator to the start of the %image
    reverse_iterator rbegin() const;
    /// Return an STL compliant reverse iterator to the end of the %image
    reverse_iterator rend() const;
    /// Return an STL compliant iterator at the point `(x, y)`
    iterator at(int x, int y) const;

    /** Return a fast STL compliant iterator to the start of the %image which must be contiguous
     *
     * @param contiguous Pixels are contiguous (must be true)
     *
     * @throws lsst::pex::exceptions::RuntimeError Argument `contiguous` is false, or the pixels are not in
     * fact contiguous
     */
    fast_iterator begin(bool contiguous) const;
    /** Return a fast STL compliant iterator to the end of the %image which must be contiguous
     *
     * @param contiguous Pixels are contiguous (must be true)
     *
     * @throws lsst::pex::exceptions::RuntimeError Argument `contiguous` is false, or the pixels are not in
     * fact contiguous
     */
    fast_iterator end(bool contiguous) const;

    /** Return an `x_iterator` to the start of the `y`'th row
     *
     * Incrementing an `x_iterator` moves it across the row
     */
    x_iterator row_begin(int y) const { return _gilView.row_begin(y); }

    /// Return an `x_iterator` to the end of the `y`'th row
    x_iterator row_end(int y) const { return _gilView.row_end(y); }

    /// Return an `x_iterator` to the point `(x, y)` in the %image
    x_iterator x_at(int x, int y) const { return _gilView.x_at(x, y); }

    /** Return an `y_iterator` to the start of the `y`'th row
     *
     * Incrementing an `y_iterator` moves it up the column
     */
    y_iterator col_begin(int x) const { return _gilView.col_begin(x); }

    /// Return an `y_iterator` to the start of the `y`'th row
    y_iterator col_end(int x) const { return _gilView.col_end(x); }

    /// Return an `y_iterator` to the point `(x, y)` in the %image
    y_iterator y_at(int x, int y) const { return _gilView.y_at(x, y); }

    /** Return an `xy_locator` at the point `(x, y)` in the %image
     *
     * Locators may be used to access a patch in an image
     */
    xy_locator xy_at(int x, int y) const { return xy_locator(_gilView.xy_at(x, y)); }
    /**
     * Set the ImageBase's origin
     *
     * The origin is usually set by the constructor, so you shouldn't need this function
     *
     * @note There are use cases (e.g. memory overlays) that may want to set these values, but
     * don't do so unless you are an Expert.
     */
    void setXY0(lsst::geom::Point2I const origin) { _origin = origin; }
    /**
     * Set the ImageBase's origin
     *
     * The origin is usually set by the constructor, so you shouldn't need this function
     *
     * @note There are use cases (e.g. memory overlays) that may want to set these values, but
     * don't do so unless you are an Expert.
     */
    void setXY0(int const x0, int const y0) { setXY0(lsst::geom::Point2I(x0, y0)); }

    lsst::geom::Box2I getBBox(ImageOrigin origin = PARENT) const {
        if (origin == PARENT) {
            return lsst::geom::Box2I(_origin, getDimensions());
        } else
            return lsst::geom::Box2I(lsst::geom::Point2I(0, 0), getDimensions());
    }

private:
    lsst::geom::Point2I _origin;
    Manager::Ptr _manager;
    _view_t _gilView;

    // oring of ImageBase in some larger image as returned to and manipulated
    // by the user

protected:
    static _view_t _allocateView(lsst::geom::Extent2I const& dimensions, Manager::Ptr& manager);
    static _view_t _makeSubView(lsst::geom::Extent2I const& dimensions, lsst::geom::Extent2I const& offset,
                                const _view_t& view);

    _view_t _getRawView() const { return _gilView; }

    inline bool isContiguous() const { return begin() + getWidth() * getHeight() == end(); }
};

template <typename PixelT>
void swap(ImageBase<PixelT>& a, ImageBase<PixelT>& b);

// Inline template definitions

template <typename PixelT>
typename ImageBase<PixelT>::Array ImageBase<PixelT>::getArray() {
    int rowStride = 1;
    PixelT * data = nullptr;
    if (getArea() > 0) {
        data = reinterpret_cast<PixelT*>(row_begin(0));
    }
    if (getHeight() > 1) {
        rowStride = reinterpret_cast<PixelT*>(row_begin(1)) - reinterpret_cast<PixelT*>(row_begin(0));
    }
    return ndarray::external(data,
                             ndarray::makeVector(getHeight(), getWidth()), ndarray::makeVector(rowStride, 1),
                             this->_manager);
}

template <typename PixelT>
typename ImageBase<PixelT>::ConstArray ImageBase<PixelT>::getArray() const {
    int rowStride = 1;
    PixelT * data = nullptr;
    if (getArea() > 0) {
        data = reinterpret_cast<PixelT*>(row_begin(0));
    }
    if (getHeight() > 1) {
        rowStride = reinterpret_cast<PixelT*>(row_begin(1)) - reinterpret_cast<PixelT*>(row_begin(0));
    }
    return ndarray::external(data,
                             ndarray::makeVector(getHeight(), getWidth()), ndarray::makeVector(rowStride, 1),
                             this->_manager);
}

}  // namespace image
}  // namespace afw
}  // namespace lsst

#endif
